
//####################################################################\\
//####################################################################\\

#define STRICT

#include <windows.h>
#include <windowsx.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#define COMMON extern

#include "..\..\cppw\standard\standard.h"
#include "..\..\cppw\standard\standarw.h"
#include "..\..\cppw\standard\grafwin.h"
#include "stackgrf.h"

  COLOR  BLACK        = RGB ( 0   , 0   , 0   );     /* dark colors */
  COLOR  BLUE         = RGB (200, 255, 200);     // = RGB ( 0   , 0   , 128 );
  COLOR  GREEN        = RGB ( 0   , 128 , 0   );
  COLOR  CYAN         = RGB ( 128 , 0   , 128 );
  COLOR  RED          = RGB ( 128 , 0   , 0   );
  COLOR  MAGENTA      = RGB ( 0   , 128 , 128 );
  COLOR  BROWN        = RGB ( 128 , 128 , 0   );
  COLOR  LIGHTGRAY    = RGB ( 192 , 192 , 192 );
  COLOR  DARKGRAY     = RGB (  96 ,  96 ,  96 );     /* light colors */
  COLOR  LIGHTBLUE    = RGB ( 127, 255, 255);     // = RGB ( 0   , 0   , 255 );
  COLOR  LIGHTGREEN   = RGB (   0, 255, 0  );
  COLOR  LIGHTCYAN    = RGB ( 255, 128, 255);
  COLOR  LIGHTRED     = RGB ( 255, 0  , 0  );
  COLOR  LIGHTMAGENTA = RGB ( 255, 128, 128);
  COLOR  YELLOW       = RGB ( 255, 255, 128);
  COLOR  LIGHTYELLOW  = RGB ( 234, 244, 157);
  COLOR  WHITE        = RGB ( 255, 255, 255);

double base;
int    iForm;
char FormatString [32];

//####################################################################\\
BOOL Get_Sutable_Range (double aMin, double aMax, int aDiv, double* bMin,
  double* bMax, int* bDiv, double* bStep, int* mnTag, int* mnOrder, double* mdOrder){
  int        nOrder;
  double     dOrder;
  int        nTag;
  double     aStep;
  double     stp;
  double     aMin1;
  double     aMax1;
  if(aMax == aMin) return FALSE;
  if(aDiv < 1) aDiv = 1;
  if(aMax < aMin){aMax1  = aMax; aMax = aMin; aMin = aMax1;}
  aStep = (aMax - aMin)/aDiv;
  nOrder = lfloor(log10(aStep) + 1.0e-12);
  dOrder = pow10(nOrder);
  nTag = nfloor(aStep/dOrder + 1.0e-12);
  if(nTag < 1) nTag = 1;
  if(nTag > 9) nTag = 9;
  switch(nTag){
    case 1: break;
    case 2: break;
    case 3:
    case 4: nTag = 4; break;
    case 5:
    case 6:
    case 7: nTag = 5; break;
    case 8:
    case 9: nTag = 1; nOrder++; break;
  }
  dOrder = pow10(nOrder);
  stp = nTag*dOrder;
  aMin1 = floor(aMin / stp)*stp;
  if(aMin1 > aMin){aMin1 = aMin1 - stp;}
  aMax1 = ceil(aMax / stp)*stp;
  if(aMax1 < aMax){aMax1 = aMax1 + stp;}
  *bDiv  = nfloor((aMax1 - aMin1)/stp + 0.1);
  *bStep = stp;
  *bMin  = aMin1;
  *bMax  = aMax1;
  *mnTag = nTag;
  *mnOrder = nOrder;
  *mdOrder = dOrder;
  return TRUE;
}

//####################################################################\\
int   PixFromIndex ( int   j, int   nSteps, int   xLeft, int   xRight )
{
   if ( j <= 0 ) return xLeft;
   if ( j >= nSteps ) return xRight;
   return ( xLeft + ((long)j * (long)( xRight - xLeft ))/(long)nSteps );
}


void SetDivision(double amin, double amax, int Nd, double* abeg, double* astep);

#pragma argsused
GRAFWIN::GRAFWIN(
   WINSTON* mws
,  char   * mGWGraphTitle = "Title"
,  PIXRECT  mGWPRG = PIXRECT()
,  TIPSRECT mGWTRC = TIPSRECT()
,  MATHRECT mGWMRD = MATHRECT()
,  PENSTYLE mGWPS  = PENSTYLE(MAGENTA)
,  COLORSCHEME mGWCS = COLORSCHEME(
                          WHITE                // Axis
                        , GREEN                // Grid
                        , LIGHTCYAN            // Frame
                        , DARKGRAY             // Fill
                        , LIGHTGRAY )          // Box
,  int      mGWStyleGrid      =    1           //
,  AXSTYLE  mGWASX  = AXSTYLE()
,  AXSTYLE  mGWASY  = AXSTYLE()
  ){
  int i;
  pws = mws;
  GWPRG = mGWPRG;
  GWTipsRectContainer = mGWTRC;
  GWMRG = mGWMRD;
  GWMRD = mGWMRD;
  for(i = 0; i < GWnPens; i++) GWPenStyle[i] = mGWPS;
  GWPenStyle[0].Color = LIGHTCYAN;
  GWPenStyle[1].Color = LIGHTYELLOW;
  GWPenStyle[2].Color = GREEN;
  GWPenStyle[3].Color = MAGENTA;
  GWPenStyle[4].Color = LIGHTCYAN;
  GWPenStyle[5].Color = LIGHTCYAN;
  GWPenStyle[6].Color = LIGHTCYAN;
  GWPenStyle[7].Color = LIGHTCYAN;
  GWColorScheme = mGWCS;
  GWStyleGrid = mGWStyleGrid;
  GWASX = mGWASX;
  GWASY = mGWASY;
//  Parameters had'nt been defined:
   GWASX.MathAnotherAxisCoord = GWMRG.GetMinX();
   GWASY.MathAnotherAxisCoord = GWMRG.GetMinY();
}

//####################################################################\\
PIXEL GRAFWIN::GWPixFromTipsX(TIPS x)
{return (TIPS)(GWPRWnd.left
   + ((long)x * (GWPRWnd.right - GWPRWnd.left)) / (long)pws->WSDefWid);
}
void GRAFWIN::GWPixFromTipsX ( PIXEL* p, TIPS* t )
{ *p = (TIPS)( GWPRWnd.left
   + ((long)(*t) * (long)(GWPRWnd.right - GWPRWnd.left)) / (long)pws->WSDefWid);
}
void GRAFWIN::GWPixFromTipsY(PIXEL* p, TIPS* t)
{ *p = (TIPS)(GWPRWnd.top
   + ((long)(*t) * (long)(GWPRWnd.bottom - GWPRWnd.top)) / (long)pws->WSDefHei);
}
void GRAFWIN::GWPixRectFromTipsRect (PIXRECT* pr, TIPSRECT* tr){
   GWPixFromTipsX ( &(pr->left)  , &(tr->left) );
   GWPixFromTipsX ( &(pr->right) , &(tr->right) );
   GWPixFromTipsY ( &(pr->top)   , &(tr->top) );
   GWPixFromTipsY ( &(pr->bottom), &(tr->bottom) );
}

PIXEL GRAFWIN::GWPixFromTipsY (TIPS y){
  return(GWPRWnd.top+(long(y)*long(GWPRWnd.bottom - GWPRWnd.top))/long(pws->WSDefHei));
}

#pragma argsused
//####################################################################\\
void GRAFWIN::GWPrepareAll(char Title[]){
  GWPrepareData();
  GWPrepareRegion(Title);
}

#pragma argsused
//####################################################################\\
void GRAFWIN::GWPrepareData(void){
   int j;
   RECT rect;
   double tmpMin, tmpMax, tmpStep;
   int    tmpDiv;
   int    nTag;
   int    nOrder;
   double   dOrder;
   int SymbWidth  = 5;
   int SymbHeight = 10;
     GetClientRect(pws->WShWnd, &rect);
     GWPRWnd.left   = rect.left;
     GWPRWnd.top    = rect.top;
     GWPRWnd.right  = rect.right;
     GWPRWnd.bottom = rect.bottom;
   GWPRC.left   = rect.left; // GWPixFromTipsX(GWTipsRectContainer.left); //???
   GWPRC.right  = rect.right; // GWPixFromTipsX(GWTipsRectContainer.right);
   GWPRC.top    = rect.top; // GWPixFromTipsY(GWTipsRectContainer.top);
   GWPRC.bottom = rect.bottom; // GWPixFromTipsY(GWTipsRectContainer.bottom);
     GWPRB.left   = GWPRC.left   + GWASX.PixBorder1;
     GWPRB.right  = GWPRC.right  - GWASX.PixBorder2;
     GWPRB.top    = GWPRC.top    + GWASY.PixBorder1;
     GWPRB.bottom = GWPRC.bottom - GWASY.PixBorder2;
   BeginFont(this->FontNumbers, SymbHeight, SymbWidth, pws->WShDC, pws->WShWnd);
     GWPRG.left   = GWPRB.left   + 4.1 * SymbWidth; // paint +4.2*
     GWPRG.right  = GWPRB.right  - 1.1 * SymbWidth;
     GWPRG.top    = GWPRB.top    + 1.1 * SymbHeight;
     GWPRG.bottom = GWPRB.bottom - 1.6 * SymbHeight;
   EndFont(pws->WShDC);
   if(GWASX.nLargeTics < 1) GWASX.nLargeTics = 1;
   Get_Sutable_Range(GWMRD.MinX, GWMRD.MaxX, GWASX.nLargeTics,
           &tmpMin, &tmpMax, &tmpDiv, &tmpStep, &nTag, &nOrder, &dOrder);
   GWMRG.MinX = tmpMin;
   GWMRG.MaxX = tmpMax;
   GWASX.MathMin = tmpMin;
   GWASX.MathMax = tmpMax;
   GWASX.nLargeTicsM = tmpDiv;
   GWASX.Step = tmpStep;
   GWASX.nTag = nTag;
   GWASX.nOrder = nOrder;
   GWASX.dOrder = dOrder;
  switch(GWASX.Style){
  case GR_AXISCENTER:
    if ( GWMRG.MinX > 0.0 )
      { j = 0; }
    else{
        if ( GWMRG.MaxX < 0.0 )
           { j = GWASX.nLargeTicsM; }
        else
           { j = nfloor ( - GWMRG.MinX / GWASX.Step + 0.1 ); }
      }
      GWASX.PixAxisCoord = PixFromIndex (j, GWASX.nLargeTicsM, GWPRG.left, GWPRG.right);
      break;
  case GR_AXISENDM:
      GWASX.PixAxisCoord = GWPRG.right + SymbWidth/2;
      break;
  case GR_AXISEND:
      GWASX.PixAxisCoord = GWPRG.right;
      break;
  case GR_AXISBEGINM:
      GWASX.PixAxisCoord = GWPRG.left - SymbWidth/2;
      break;
  case GR_AXISBEGIN:
    default:
    GWASX.PixAxisCoord = GWPRG.left;
    break;
  }
  if(GWASY.nLargeTics < 1) GWASY.nLargeTics = 1;
  Get_Sutable_Range(GWMRD.MinY, GWMRD.MaxY, GWASY.nLargeTics,
           &tmpMin, &tmpMax, &tmpDiv, &tmpStep, &nTag, &nOrder, &dOrder);
   GWMRG.MinY = tmpMin;
   GWMRG.MaxY = tmpMax;
   GWASY.MathMin = tmpMin;
   GWASY.MathMax = tmpMax;
   GWASY.nLargeTicsM = tmpDiv;
   GWASY.Step = tmpStep;
   GWASY.nTag = nTag;
   GWASY.nOrder = nOrder;
   GWASY.dOrder = dOrder;
  switch ( GWASY.Style ){
  case GR_AXISCENTER:
    if ( GWMRG.MinY > 0.0 )
       { j = 0; }
    else{
      if ( GWMRG.MaxY < 0.0 )
         { j = GWASY.nLargeTicsM; }
      else
         { j = nfloor ( - GWMRG.MinY / GWASY.Step + 0.1 );}
    }
    GWASY.PixAxisCoord = PixFromIndex(j, GWASY.nLargeTicsM, GWPRG.bottom, GWPRG.top);
        break;
     case GR_AXISENDM:
        GWASY.PixAxisCoord = GWPRG.top - SymbHeight/2;
        break;
     case GR_AXISEND:
        GWASY.PixAxisCoord = GWPRG.top;
        break;
     case GR_AXISBEGINM:
        GWASY.PixAxisCoord = GWPRG.bottom + SymbHeight/2;
        break;
     case GR_AXISBEGIN:
     default:
        GWASY.PixAxisCoord = GWPRG.bottom;
        break;
  }
  GWASY.Scale = ( GWPRG.bottom - GWPRG.top  )/( GWMRG.MaxY - GWMRG.MinY );
  GWASX.Scale = ( GWPRG.right  - GWPRG.left )/( GWMRG.MaxX - GWMRG.MinX );
  GWASX.MathAnotherAxisCoord = GWMathFromPixX (GWASX.PixAxisCoord);
  GWASY.MathAnotherAxisCoord = GWMathFromPixY (GWASY.PixAxisCoord);
}

#pragma argsused
//####################################################################\\
void GRAFWIN::GWPrepareRegion (char * Title){
  int SymbWidth  = 5;
  int SymbHeight = 10;
  PIXRECT r;
  SetBkMode(pws->WShDC, TRANSPARENT);
  BeginBrush (PS_SOLID, GWColorScheme.Frame, pws->WShDC, pws->WShWnd);
  BeginPen (PS_SOLID, 1, GWColorScheme.Frame, pws->WShDC, pws->WShWnd);
     r.left   = GWPRWnd.left;
     r.top    = GWPRWnd.top;
     r.right  = GWPRC.left+1;
     r.bottom = GWPRWnd.bottom;
     GWRectangle ( &r );
     r.left   = GWPRC.left;
     r.top    = GWPRWnd.top;
     r.right  = GWPRC.right;
     r.bottom = GWPRC.top+1;
     GWRectangle (&r);
     r.left   = GWPRC.left;
     r.top    = GWPRC.bottom;
     r.right  = GWPRC.right;
     r.bottom = GWPRWnd.bottom;
     GWRectangle (&r);
     r.left   = GWPRC.right;
     r.top    = GWPRWnd.top;
     r.right  = GWPRWnd.right;
     r.bottom = GWPRWnd.bottom;
     GWRectangle (&r);
   EndPen   (pws->WShDC);
   EndBrush (pws->WShDC);
   BeginBrush(PS_SOLID, GWColorScheme.Box, pws->WShDC, pws->WShWnd);
   BeginPen(PS_SOLID, 1, GWColorScheme.Box, pws->WShDC, pws->WShWnd);
     r.left   = GWPRC.left;
     r.top    = GWPRC.top;
     r.right  = GWPRG.left+1;
     r.bottom = GWPRC.bottom;
     GWRectangle ( &r );
     r.left   = GWPRG.left;
     r.top    = GWPRC.top;
     r.right  = GWPRG.right;
     r.bottom = GWPRG.top+1;
     GWRectangle ( &r );
     r.left   = GWPRG.left;
     r.top    = GWPRG.bottom;
     r.right  = GWPRG.right;
     r.bottom = GWPRC.bottom;
     GWRectangle ( &r );
     r.left   = GWPRG.right;
     r.top    = GWPRC.top;
     r.right  = GWPRC.right;
     r.bottom = GWPRC.bottom;
     GWRectangle ( &r );
   EndPen   ( pws->WShDC );
   EndBrush ( pws->WShDC );
   BeginBrush ( PS_SOLID, GWColorScheme.Fill, pws->WShDC, pws->WShWnd );
     r.left   = GWPRG.left;
     r.top    = GWPRG.top;
     r.right  = GWPRG.right;
     r.bottom = GWPRG.bottom+1;
     GWRectangle ( &r ); // & GWPRG );
   EndBrush ( pws->WShDC );
  if (GWASX.Style) GWAxisX();
  if (GWASY.Style) GWAxisY();
  BeginFont (this->FontCaps, SymbHeight, SymbWidth, pws->WShDC, pws->WShWnd);
    SetTextAlign (pws->WShDC, TA_RIGHT | TA_BASELINE);
    TextOut (pws->WShDC, GWPRWnd.right-SymbWidth, SymbHeight+2, Title, strlen(Title));
  EndFont(pws->WShDC);
}

//####################################################################\\
void SelectFormat(char* buf, double aMin, double aMax, double dOrder, int order){
  base = 1.0;
  iForm = order;
  double dMax = dmax((aMin / dOrder), (aMax / dOrder));
//        if ( order == -3 && aMin >= 0.0 && aMax < 9999.999 )
//                              { strcpy ( buf, "%5.3f"); }
//   else
        if ( order == -2 && dMax < 999.999 )
                              { strcpy ( buf, "%5.2f"); }
   else if ( order == -1 && dMax < 999.999 )
                              { strcpy ( buf, "%5.1f"); }
   else if ( order ==  0 && dMax < 9999.999 )
                              { strcpy ( buf,   "%5d"); }
   else if ( order ==  1 && dMax < 999.999 )
                              { strcpy ( buf,   "%5d"); }
   else if ( order ==  2 && dMax <  99.999 )
                              { strcpy ( buf,   "%5d"); }
   else if ( order ==  3 && dMax <   9.999 )
                              { strcpy ( buf,   "%5d"); }
   else{
//      strcpy ( buf, "%#-10.2lg");
      strcpy ( buf, "%5d");
      iForm = 10;
      base = dOrder;
   }
}

//####################################################################\\
void FormatLabel(double xt, char * buf, char * FormatString){
   int j;
   double r;
   switch ( iForm )
   {
   case -3:
   case -2:
   case -1:
   default:
      sprintf( buf, FormatString, xt / base );
      break;
   case 0:
   case 1:
   case 2:
   case 3:
      sprintf( buf, FormatString, nfloor (xt / base + 0.1) );
   case 10:
      r = xt / base;
      if ( fabs ( r ) < 9999.9 )
           sprintf( buf, FormatString, nfloor (r+0.1) );
      else strcpy ( buf,"***");
      break;
   } // switch
   for( j = strlen( buf )-1; j>0; j-- )
   {
      if( buf[j] != ' ' ) break;
   }
   if ( j > 1 )
   {
      if ( buf[j] == '0' && buf[j-2] == '.'){ j--;}
      if ( buf[j] == '.'){ buf[j+1] = '0'; j++;}
   }
   buf[j+1] = '\0';
}

#pragma argsused
//####################################################################\\
void GRAFWIN::GWAxisX(void){
  FONTMODEL mFontNameX    = FontNames;
  FONTMODEL mFontNumbersX = FontNumbers;
  int X,Y;
//  POINT point;
  int SymbWidth, SymbHeight;
  double xt, dxt, xs;
  int i, j, order;
  char buf[20];
// Axis X Line:
  BeginPen ( PS_SOLID, 1, GWColorScheme.Axis, pws->WShDC, pws->WShWnd );
  GWLinePix(GWPRG.left, GWASY.PixAxisCoord, GWPRG.right, GWASY.PixAxisCoord);
//  EndPen (pws->WShDC);
//  BeginPen ( PS_SOLID, 1, WHITE, pws->WShDC, pws->WShWnd );
  X = GWPRG.right;
  Y = GWASY.PixAxisCoord;
  GWLinePix( X-6, Y+4, X, Y );
  GWLinePix( X-6, Y-4, X, Y );
  EndPen (pws->WShDC);
//  Axis X Name:
  BeginFont (mFontNameX, SymbHeight, SymbWidth, pws->WShDC, pws->WShWnd);
  SetTextAlign (pws->WShDC, TA_RIGHT | TA_BASELINE);
  GWMoveRelPix (-SymbWidth, SymbHeight*1.3);
  SetTextColor (pws->WShDC, GWColorScheme.Axis);
  GWTextOut (GWASX.Name);
  EndFont (pws->WShDC);
  BeginFont(mFontNumbersX, SymbHeight, SymbWidth, pws->WShDC, pws->WShWnd);
  if(GWASX.nLargeTicsM != 0){
     order = GWASX.nOrder;
     SelectFormat(FormatString, GWASX.MathMin, GWASX.MathMax, GWASX.dOrder, order);
     dxt = GWASX.Step;
     if(GWASX.nSmallTics > 0) dxt = GWASX.Step / GWASX.nSmallTics;
     SetTextAlign (pws->WShDC, TA_CENTER | TA_BASELINE );
     BeginPen(PS_SOLID, 3, GWColorScheme.Box, pws->WShDC, pws->WShWnd );
     for(i = 0; i <= GWASX.nLargeTicsM; i++){
        xt = GWMRG.MinX + i * GWASX.Step;
        if( fabs( xt ) < 0.001 * fabs( GWASX.Step ) )
           { xt = 0.0;}
        GWMoveToMath( xt, GWASY.MathAnotherAxisCoord );
        GWLineRelPix( 0, -6 );
// Axis X: Grid lines:
        if(GWStyleGrid){
           BeginPen(PS_DOT, 1, GWColorScheme.Grid, pws->WShDC, pws->WShWnd);
           if(fabs(xt - GWASX.MathAnotherAxisCoord) > 0.001*fabs(GWASX.Step)){
              GWLineMath(xt, GWMRG.MinY, xt, GWMRG.MaxY);
           }
           EndPen ( pws->WShDC );
        }
// Axis X Labels:
//   sprintf( buf, FormatString, xt / pow10( order ) );
        FormatLabel(xt, buf, FormatString);
        GWMoveToMath(xt, GWASY.MathAnotherAxisCoord);
        GWMoveRelPix( -SymbWidth, SymbHeight*1.0 );
        SetTextColor(pws->WShDC, GWColorScheme.Axis);
        SetTextAlign(pws->WShDC, TA_CENTER | TA_BASELINE);
        if((i != GWASX.nLargeTicsM) || (order == 0))
           {GWTextOut(buf);}
// Axis X: Small Tics:
        xs = xt;
        BeginPen ( PS_SOLID, 3, GWColorScheme.Box, pws->WShDC, pws->WShWnd );
        for( j = 1; j < GWASX.nSmallTics; j++ )
        {
           xs += dxt;
           if( ( xs <= GWMRG.MinX ) || ( xs >= GWMRG.MaxX ) ) break;
           GWMoveToMath( xs, GWASY.MathAnotherAxisCoord );
           GWLineRelPix( 0, -3 );
        } // j;
        EndPen ( pws->WShDC );
     } // i;
     EndPen ( pws->WShDC );
// Axis X: Order:
     if( iForm == 10 )
     {
        GWMoveToMath(GWMRG.MaxX, GWASY.MathAnotherAxisCoord);
        GWMoveRelPix(-SymbWidth*2.3, -SymbHeight);
        SetTextAlign(pws->WShDC, TA_RIGHT | TA_BASELINE);
        GWTextOut("*10");
        sprintf( buf, "%-3d", order );
        for( j = strlen( buf )-1; j>0; j-- )
        {
           if( buf[j] != ' ' ) break;
        }
               buf[j+1] = '\0';
        SetTextAlign(pws->WShDC, TA_LEFT | TA_BASELINE);
        GWMoveRelPix(SymbWidth*0.8, -SymbHeight*0.7);
        GWTextOut(buf);
     }
  }
  EndFont (pws->WShDC);
}

//####################################################################\\
void GRAFWIN::GWAxisY (void){
  FONTMODEL mFontNameY = FontNames;
  FONTMODEL mFontNumbersY = FontNumbers;
  int X,Y;
  POINT point;
  double yt, dyt, ys;
  int i,j,order;
  int SymbWidth, SymbHeight ;
  char buf[20];
  BeginPen(PS_SOLID, 1, GWColorScheme.Axis, pws->WShDC, pws->WShWnd);
  GWLineMath(GWASX.MathAnotherAxisCoord, GWMRG.MinY, GWASX.MathAnotherAxisCoord, GWMRG.MaxY);
  GWMoveToMath(GWASX.MathAnotherAxisCoord, GWMRG.MaxY);
  GetCurrentPositionEx(pws->WShDC, &point);
  EndPen(pws->WShDC);
  BeginPen(PS_SOLID, 1, WHITE, pws->WShDC, pws->WShWnd);
    X = point.x;
    Y = point.y;
    GWLinePix (X-4, Y+6, X, Y);
    GWLinePix (X+4, Y+6, X, Y);
  EndPen (pws->WShDC);
  BeginFont (mFontNameY, SymbHeight, SymbWidth, pws->WShDC, pws->WShWnd);
    SetTextAlign (pws->WShDC, TA_LEFT | TA_BASELINE);
    GWMoveRelPix (1.8*SymbWidth, -SymbHeight/3);
    SetTextColor (pws->WShDC, GWColorScheme.Axis);
    GWTextOut (GWASY.Name);
  EndFont (pws->WShDC);
  BeginFont (mFontNumbersY, SymbHeight, SymbWidth, pws->WShDC, pws->WShWnd);
  if (GWASY.nLargeTicsM != 0){
    order = GWASY.nOrder;
    SelectFormat(FormatString, GWASY.MathMin, GWASY.MathMax, GWASY.dOrder, order);
    dyt = GWASY.Step;
    if ( GWASY.nSmallTics > 1) { dyt = GWASY.Step / GWASY.nSmallTics;}
    for(i = 0; i <= GWASY.nLargeTicsM; i++){
      yt = GWMRG.MinY + i * GWASY.Step;
      if(fabs(yt) < 0.001 * fabs(GWASY.Step))
        {yt = 0.0;}
// Axis Y: Large Tics:
      GWMoveToMath (GWASX.MathAnotherAxisCoord, yt);
      GWLineRelPix(5, 0);
      if (GWStyleGrid){
        BeginPen (PS_DOT, 1, GWColorScheme.Grid, pws->WShDC, pws->WShWnd);
        if(fabs(yt - GWASY.MathAnotherAxisCoord) > 0.001 * fabs(GWASY.Step)){
          GWLineMath (GWMRG.MinX, yt, GWMRG.MaxX, yt);
        }
        EndPen (pws->WShDC);
      }
//  Axis Y Labels:
      FormatLabel (yt, buf, FormatString);
      GWMoveToMath (GWASX.MathAnotherAxisCoord, yt);
      SetTextColor (pws->WShDC, GWColorScheme.Axis);
      SetTextAlign (pws->WShDC, TA_RIGHT | TA_BASELINE); // paint TA_RIGHT
      GWMoveRelPix (-0.2*SymbWidth, SymbHeight/3);
      if ((i != GWASY.nLargeTicsM) || (order == 0)) GWTextOut(buf);
// Axis Y Small Tic:
      ys = yt;
      for(j = 1; j < GWASY.nSmallTics; j++){
        ys += dyt;
        if( ( ys < GWMRG.MinY ) || ( ys > GWMRG.MaxY ) ) break;
        GWMoveToMath( GWASX.MathAnotherAxisCoord, ys );
        GWLineRelPix( 3, 0 );
      }
    } //  i;
// Axis Y Order Mark:
    if(iForm == 10){
      GWMoveToMath( GWASX.MathAnotherAxisCoord, GWMRG.MaxY );
      GWMoveRelPix( -1.2*SymbWidth, SymbHeight/5 );
      SetTextAlign( pws->WShDC, TA_RIGHT | TA_BASELINE );
      GWTextOut("*10");
      sprintf( buf, "%-3d", order );
      for( j = strlen( buf )-1; j>0; j-- )
         { if( buf[j] != ' ' ) break; }
      buf[j+1] = '\0';
      SetTextAlign( pws->WShDC, TA_LEFT | TA_BASELINE );
      GWMoveRelPix ( SymbWidth/5, -SymbHeight/2 );
      GWTextOut(buf);
    }
  }
  EndFont (pws->WShDC);
}

//####################################################################\\
void DummyAutoFontScale ( FONTMODEL &fontCurrent, HDC hdc, HWND hWnd){
  int SymbHeight, SymbWidth;
  int ClientPixX, ClientPixY;
  TEXTMETRIC textMetrics;
  RECT rect;
    GetClientRect (hWnd, &rect);
    ClientPixX = rect.right - rect.left;
    ClientPixY = rect.bottom - rect.top;
  GetTextMetrics (hdc, &textMetrics);
//  SymbHeight = textMetrics.tmHeight;
//  SymbWidth  = textMetrics.tmAveCharWidth;
    int optFontSizeX =  ( 800 + ClientPixX/100 )/ fontCurrent.CoefFontX;
    int optFontSizeY =  ( 600 + ClientPixY/100 )/ fontCurrent.CoefFontY;
    SymbWidth  = optFontSizeX;
    if(SymbWidth  < 6) SymbWidth = 6;
  fontCurrent.Width = SymbWidth;
    SymbHeight = optFontSizeY;
  if(SymbHeight < 10) SymbHeight = 10;
  fontCurrent.Height = SymbHeight;
}

//####################################################################\\
void TextOutMy (HDC hdc, char buf[]){
   POINT point;
   GetCurrentPositionEx (hdc, &point);
   TextOut ( hdc, point.x, point.y, buf,strlen(buf) );
}

//####################################################################\\
void GRAFWIN::GWTextOut (char buf[]){
   static POINT point;
   GWGetCurrentPositionEx (&point);
   TextOut ( pws->WShDC, point.x, point.y, buf, strlen(buf));
}

//####################################################################\\
PIXEL GRAFWIN::GWPixFromMathX(double a){
  return(GWPRG.left + floor(( a - GWMRG.MinX ) * GWASX.Scale));
}
PIXEL GRAFWIN::GWPixFromMathY(double a){
  return(GWPRG.bottom - ceil((a - GWMRG.MinY)*GWASY.Scale));
}
double GRAFWIN::GWMathFromPixX( PIXEL x ){
   return ( GWMRG.MinX + ( double( x - GWPRG.left ) / GWASX.Scale ) );
}
double GRAFWIN::GWMathFromPixY(PIXEL y){
  return(GWMRG.MinY - (double(y - GWPRG.bottom)/GWASY.Scale));
}

//####################################################################\\
void Calc_MinMax(double* pArray, int nJump, int nValues, double* yMin, double* yMax){
  int i, index = nJump;
  *yMax = *yMin = pArray[0];
  for(i = 1; i < nValues; i++)
  {
     *yMax = dmax(*yMax, pArray[index]);
     *yMin = dmin(*yMin, pArray[index]);
     index += nJump;
  }
}

//####################################################################\\
void Calc_MinMaxm(double* pArray, int na, int nbp1, double* yMin, double* yMax){
  int i;
  *yMax = *yMin = pArray[na];
  for(i = na+1; i < nbp1; i++){
     *yMax = dmax(*yMax, pArray[i]);
     *yMin = dmin(*yMin, pArray[i]);
  }
}

//####################################################################\\
void GRAFWIN::GWMarkLine(int x, int y, int wid, int hig, char* Text, int iPen){
   PIXRECT rect;
   GWBeginBrush(PS_SOLID, clr[iPen]); // GetPenColor(iPen));
   rect.left   = x;
   rect.top    = y;
   rect.right  = x+wid;
   rect.bottom = y+hig;
   GWRectangle (&rect);
   GWMoveTo(rect.right+5, rect.bottom);
   GWSetTextAlign(TA_LEFT | TA_BASELINE);
   GWTextOut(Text);
   GWEndBrush();
}

//####################################################################\\
//  OLD VERSION!!!
void GRAFWIN::GWPaint_yCurveMath (
            double * Y         // double * Y
         , double   xBeg      // double   xBeg
         , double   xStep     // double   xStep
         , int      nPoints   // int      nPoints
           )
//  OLD VERSION!!!
{
  int i;
  double xNew, yNew;
  if ( nPoints < 2) return;
  POINT FAR * pt = new POINT [ nPoints ];
  if ( !pt ) return;
    BeginPen ( GWPenStyle[0].Style, GWPenStyle[0].Width, GWPenStyle[0].Color,
               pws->WShDC, pws->WShWnd );
    pt[0].x = GWPixFromMathX ( xBeg );
    pt[0].y = GWPixFromMathY ( Y[0] );
    for ( i = 1; i < nPoints; i++ )
    {
       xNew = xBeg + xStep * i;
       yNew = Y [ i ];
       pt[i].x = GWPixFromMathX( xNew );
       pt[i].y = GWPixFromMathY( yNew );
    }
    Polyline ( pws->WShDC, pt, nPoints );
    delete pt;
    EndPen ( pws->WShDC );
}

//####################################################################\\
void GRAFWIN::GWPaint_xyPointsMath (
            double * X         // double * X
         , double * Y         // double * Y
         , POINT  * pt        // POINT * P
         , int      nPoints   // int      nPoints
         , int      iPen      //
         , int      Enclosed
           )
{
  int i;
  if ( nPoints < 2) return;
  if ( !pt ) return;
    BeginPen ( GWPenStyle[iPen].Style, GWPenStyle[iPen].Width, GWPenStyle[iPen].Color,
               pws->WShDC, pws->WShWnd );
    for ( i = 0; i < nPoints; i++ )
    {
       pt[i].x = GWPixFromMathX( X[i] );
       pt[i].y = GWPixFromMathY( Y[i] );
    }
  i = nPoints;
  if ( Enclosed )
  {
    pt[i].x = GWPixFromMathX( X[0] );
    pt[i].y = GWPixFromMathY( Y[0] );
  }
  Polyline ( pws->WShDC, pt, nPoints );
  EndPen ( pws->WShDC );
}

//####################################################################\\
void GRAFWIN::GWPaint_xyCurveMath(double* X, double* Y, int nPoints, int iPen)
{
  GWPaint_xyCurveMathm(X, Y, 0, nPoints, iPen);
}

//####################################################################\\
void GRAFWIN::GWPaint_xyCurveMathm(double* X, double* Y, int na, int nbp1, int iPen){
  int i;
  PIXEL px1, py1, px2, py2;
  if (nbp1-na < 2) return;
  BeginPen(GWPenStyle[iPen].Style, GWPenStyle[iPen].Width, clr[iPen], pws->WShDC, pws->WShWnd);
  px1 = GWPixFromMathX (X[na]);
  py1 = GWPixFromMathY (Y[na]);
  for(i = na+1; i < nbp1; i++){
     px2 = GWPixFromMathX(X[i]);
     py2 = GWPixFromMathY(Y[i]);
     GWLinePix(px1, py1, px2, py2);
     px1 = px2;
     py1 = py2;
  }
  EndPen(pws->WShDC);
}

//####################################################################\\
void GRAFWIN::GWLineMath(double x1, double y1, double x2, double y2){
   int i1, i2, j1, j2;
   i1 = GWPixFromMathX(x1);
   j1 = GWPixFromMathY(y1);
   i2 = GWPixFromMathX(x2);
   j2 = GWPixFromMathY(y2);
   GWLinePix(i1, j1, i2, j2);
}

//####################################################################\\
void GRAFWIN::GWLinePix ( int x1, int y1, int x2, int y2 )
{
   MoveToEx ( pws->WShDC, x1, y1, NULL );
   LineTo ( pws->WShDC, x2, y2 );
}

//####################################################################\\
void GRAFWIN::GWLineRelPix ( int dx, int dy ){
  POINT point;
  GetCurrentPositionEx ( pws->WShDC, &point );
  LineTo ( pws->WShDC, point.x+dx, point.y+dy);
}

//####################################################################\\
void GRAFWIN::GWMoveRelPix ( int dx, int dy )
{
  POINT point;
  GetCurrentPositionEx ( pws->WShDC, & point );
  MoveToEx ( pws->WShDC, point.x+dx, point.y+dy, NULL);
}

//####################################################################\\
void GRAFWIN::GWMoveTo( int x, int y )
   {MoveToEx( pws->WShDC, x, y, NULL);}

//####################################################################\\
void GRAFWIN::GWMoveToMath( double x, double y )
   {MoveToEx( pws->WShDC, GWPixFromMathX(x), GWPixFromMathY(y), NULL);}

//####################################################################\\
void GRAFWIN::GWMoveToMath01( double ax, double ay )
{
   GWMoveToMath(
      GWGetMathMinXG() + ax * (GWGetMathMaxXG() - GWGetMathMinXG()),
      GWGetMathMinYG() + ay * (GWGetMathMaxYG() - GWGetMathMinYG()));
}

HBRUSH GRAFWIN::GWBeginBrush ( int style, DWORD color )
   {return BeginBrush (style, color, pws->WShDC, pws->WShWnd );}
void GRAFWIN::GWEndBrush (void){ EndBrush ( pws->WShDC );}

//####################################################################\\
void GRAFWIN::GWPaint_yGraph (
        double * var
     , double   xMin
     , double   xMax
     , double   xStep
     , int      nPointsAll
     , int      nCurves
     , char     Title[]
       )
{
  int ig;
  if ( nCurves < 1 ) return;
  int nPointsOne = nPointsAll / nCurves;
  Calc_MinMax( var, 1, nPointsAll, & GWMRD.MinY, & GWMRD.MaxY );
  if ( GWMRD.MaxY == GWMRD.MinY )
  {
    if ( GWMRD.MinY == 0.0 )
       GWMRD.MaxY = 1.0e-9;
    else
    {
      GWMRD.MaxY += 1.0e-9 * ( GWMRD.MaxY- GWMRD.MinY );
      if ( GWMRD.MinY == GWMRD.MaxY )
           GWMRD.MaxY += 1.0e-9;
    }
  }
  if ( GWMRD.MinY > 0.0 ) GWMRD.MinY = 0.0;
  if ( GWMRD.MaxY < 0.0 ) GWMRD.MaxY = 0.0;
//   GWSetMathAxisY (0.0);
//   GWSetMathAxisY ( GWMRD.MinY );
//   GWMRD.MaxY += 0.1 * ( GWMRD.MaxY - GWMRD.MinY );
   GWMRD.MinX = xMin;
   GWMRD.MaxX = xMax;
   GWPrepareAll ( Title );
   for ( ig = 0; ig < nCurves; ig++ )
   {
      GWPaint_yCurveMath (
            &var [ ig * nPointsOne ] // double * y
         , xMin            // double   xBeg
         , xStep           // double   xStep
         , nPointsOne      // int      nPoints
           );
   }
}

//####################################################################\\
void GRAFWIN::GWPrepare_xyMath (
        double * xPoints
     , double * yPoints
     , int      nPoints
       )
{
   Calc_MinMax( xPoints, 1, nPoints, & GWMRD.MinX, & GWMRD.MaxX );
   Calc_MinMax( yPoints, 1, nPoints, & GWMRD.MinY, & GWMRD.MaxY );
   if ( GWMRD.MaxX == GWMRD.MinX )
   {
      GWMRD.MaxX += 0.5;
      GWMRD.MinX -= 0.5;
   }
   if ( GWMRD.MaxY == GWMRD.MinY )
   {
     if ( GWMRD.MinY == 0.0 )
        GWMRD.MaxY = 1.0e-9;
     else
     {
        GWMRD.MaxY += 1.0e-9 * ( GWMRD.MaxY- GWMRD.MinY );
        if ( GWMRD.MinY == GWMRD.MaxY )
           GWMRD.MaxY += 1.0e-9;
     }
   }
   if ( GWMRD.MinY > 0.0 ) GWMRD.MinY = 0.0;
   if ( GWMRD.MaxY < 0.0 ) GWMRD.MaxY = 0.0;
}

//####################################################################\\
void GRAFWIN::GWPaint_xyGraph (
        double * xPoints
      , double * yPoints
      , int      nPoints
      , char     Title[])
{
   GWPrepare_xyMath ( xPoints, yPoints, nPoints );
   GWPrepareAll ( Title );
   GWPaint_xyCurveMath (xPoints, yPoints, nPoints, 0);
}


REFRESHLIMITSWIN::REFRESHLIMITSWIN(int n)
{
    nLiw = n;
    IsRefrLimits = new int[nLiw];
    int i;
    for ( i = 0 ; i < nLiw ; i++ ) IsRefrLimits[i]=TRUE;
}
REFRESHLIMITSWIN::~REFRESHLIMITSWIN()
{
    delete[] IsRefrLimits;
}
int REFRESHLIMITSWIN::ChkIndx(int& i)
  {if (i < 0) i=0; if (i >= nLiw) i=nLiw-1;return TRUE;}
int REFRESHLIMITSWIN::ToRefreshLimits()
{
  for ( int i = 0 ; i < nLiw ; i++ ) ToRefreshLimits(i);
  return TRUE;
}

void GWIN::DestroyAll(void)
{
  for( int i = 1; i < nOfGraphics; i++ )
    ppWinSton[i]->GWDestroy();
}
void GWIN::CreateAll(HWND hWndMain, HINSTANCE hInst){
  for( int i = 1; i < nOfGraphics; i++ ){
    if(ppWinSton[i]->GWState == GWS_CLOZED)
      ppWinSton[i]->WSCreateWindow(hWndMain, hInst);
  }
}

void GWIN::ShowAll(int nShow){
  for(int i = 1; i < nOfGraphics; i++)
    ppWinSton[i]->WSShowWindow(nShow);
}

void GWIN::ShowSet(int na, int nb, int nShow){
  for(int i = na; i <= nb; i++)
    ppWinSton[i]->WSShowWindow(nShow);
}

int GWIN::InvalWins(){
  for(int i = 1; i < nOfGraphics; i++) ppWinSton[i]->GWInvalidateRgnNT();
  return FALSE;
}

int GWIN::GWSetWinsPixPos(){
  for (int i=1; i<nOfGraphics; i++)
    ppWinSton[i]->WSSetWinPixPos(UserRect);
  return FALSE;
}



//####################################################################\\
//####################################################################\\
//####################################################################\\

//??? to standard!
int dminmax(double *dfMin, double *dfMax, PDFUNCLL pfll, int nya, int numy, int nxa, int numx)
{
  int i,j;
  *dfMax = *dfMin = pfll(nya,nxa);
  if ( numx < 1 ) return TRUE;
  if ( numy < 1 ) return TRUE;
  int nybp1 = nya + numy;
  int nxbp1 = nxa + numx;
  for (i = nya; i < nybp1; i++)
    for (j = nxa; j < nxbp1; j++)
    if (*dfMin > pfll(i,j))
        *dfMin = pfll(i,j);
    else
      if (*dfMax < pfll(i,j))
          *dfMax = pfll(i,j);
  return FALSE;
}
int dminmax(double *dfMin, double *dfMax, PDFUNCLL pfll, int numy, int numx)
{
  return dminmax(dfMin, dfMax, pfll, 0, numy, 0, numx);
}

//####################################################################\\
int dminmax(double *dfMin, double *dfMax, PDFUNCL pfl,
            int nPoints, int nLowIndex=0)
{
  if ( nPoints<1 || pfl==NULL )
  {
    *dfMax = *dfMin = 0.0;
    return TRUE;
  }
  int i;
  int mi = nLowIndex+nPoints;
  *dfMax = *dfMin = pfl(nLowIndex);
  if ( nPoints == 1 ) return FALSE;
  for (i = nLowIndex; i < mi; i++)
    if (*dfMin > pfl(i))
      *dfMin = pfl(i);
    else
      if (*dfMax < pfl(i))
          *dfMax = pfl(i);
  return FALSE;
}

double y_ang = M_PI/180.0*45.0;

double cos_y_ang;
double sin_y_ang;

//####################################################################\\
inline void Projection(double x, double y, double z, double *px, double *py)
{
   *px = x + y*cos_y_ang;
   *py = z + y*sin_y_ang;
}

//####################################################################\\
void WinDrawSurface(HDC wcmhdc, RECT *wcmr, DOUBLE2DM *srf, int mode)
{
  int nx, ny;
  int PixLenX,PixLenY;
  double min_x, min_y, max_x, max_y, min_z, max_z;
  double x,y;
  double kx,ky,kz;
  int i,j;
  nx = srf->GetDimX();
  ny = srf->GetDimY();
  if (!nx) return;
  cos_y_ang = cos(y_ang);
  sin_y_ang = sin(y_ang);
  PixLenX = wcmr->right  - wcmr->left;
  PixLenY = wcmr->bottom - wcmr->top;
//   Projection(nx-1, ny-1, 0, &max_x, &max_y);
//???  kx = PixLenX / max_x;
//???  ky = PixLenY / max_y;
  srf->D2DMinMax(&min_z,&max_z);
  Projection(0, 0, (*srf)[0][0], &min_x, &min_y);
  max_x = min_x; max_y = min_y;
  for (i = 0; i < ny; i++)
    for (j = 0; j < nx; j++)
    {
      Projection(j, i, /*(*srf)[i][j]-min_z*/0, &x, &y);
      if (x < min_x)      min_x = x;
      else if (x > max_x) max_x = x;
      if (y < min_y)      min_y = y;
      else if (y > max_y) max_y = y;
    }
  kx = PixLenX / max_x;
  ky = PixLenY / max_y;
  if (max_z == min_z) kz = 0;
  else                kz = PixLenY/2/(max_z-min_z)/ky;
  switch (mode)
  {
  case wcmLINE:
  for (i = ny-1; i >= 0; i--)
  {
    Projection(0, i*ky, kz*((*srf)[i][0]-min_z), &x, &y);
    MoveToEx(wcmhdc,wcmr->left + x,wcmr->bottom - y,NULL);
    for (j = 1; j < nx; j++)
     {
      Projection(j*kx, i*ky, kz*((*srf)[i][j]-min_z), &x, &y);
      LineTo(wcmhdc,wcmr->left + x,wcmr->bottom - y);
     }
  }
  for (j = nx-1; j >= 0; j--)
  {
    Projection(j*kx, 0, kz*((*srf)[0][j]-min_z), &x, &y);
    MoveToEx(wcmhdc,wcmr->left + x,wcmr->bottom - y,NULL);
    for (i = 1; i < ny; i++)
    {
      Projection(j*kx, i*ky, kz*((*srf)[i][j]-min_z), &x, &y);
      LineTo(wcmhdc,wcmr->left + x,wcmr->bottom - y);
    }
  }
  break;
  case wcmBRUSH:
  {
  POINT pnt[4];
  HBRUSH obr,bbr,wbr;
  int norm_y;
   bbr = CreateSolidBrush(RGB(127,127,127));
   wbr = CreateSolidBrush(RGB(255,255,255));
   obr = (HBRUSH)SelectObject(wcmhdc,wbr);

   for (i = ny - 1; i > 0; i--)
   {
     Projection((nx-1), i, kz*((*srf)[i][nx-1]-min_z), &x, &y);
     pnt[1].x = wcmr->left + x*kx; pnt[1].y = wcmr->bottom - y*ky;
     Projection((nx-1), (i-1), kz*((*srf)[i][nx-1]-min_z), &x, &y);
     pnt[2].x = wcmr->left + x*kx; pnt[2].y = wcmr->bottom - y*ky;
     for (j = nx - 2; j >= 0; j--)
     {
       pnt[0].x = pnt[1].x; pnt[0].y = pnt[1].y;
       pnt[3].x = pnt[2].x; pnt[3].y = pnt[2].y;
       Projection(j, i, kz*((*srf)[i][j]-min_z), &x, &y);
       pnt[1].x = wcmr->left + x*kx; pnt[1].y = wcmr->bottom - y*ky;
       Projection(j, (i-1), kz*((*srf)[i-1][j]-min_z), &x, &y);
       pnt[2].x = wcmr->left + x*kx; pnt[2].y = wcmr->bottom - y*ky;
       //Norm.y := -((p2.x - p1.x)*(p3.z - p2.z) - (p3.x - p2.x)*(p2.z - p1.z));*/
       norm_y = (pnt[1].x-pnt[0].x)*(pnt[2].y-pnt[1].y) -
                (pnt[2].x - pnt[1].x)*(pnt[1].y - pnt[0].y);
       if (norm_y < 0)
          SelectObject(wcmhdc,wbr);
       else
          SelectObject(wcmhdc,bbr);
       Polygon(wcmhdc,pnt,4);
     }
   }
   SelectObject(wcmhdc,obr);
   DeleteObject(bbr);
   DeleteObject(wbr);
  }
  break;
  }
}

//####################################################################\\
int WINCOLORMAP::PaintSurface2Da ( PDFUNCLL pfll, int ndy, int ndx,
                                  double dfMin, double dfMax )
{
  SetNY(ndy);
  SetNX(ndx);
  SetfMin(dfMin);
  SetfMax(dfMax);
  SetpFunc(pfll);
  return PaintSurface2Da();
}

//####################################################################\\
// New: modified to DFLL instead of  DOUBLE2DM:
int WINCOLORMAP::PaintSurface2Da()
{
  int m, n, m1, n1;
  int mx, my;
  int PixLenX,PixLenY;
  double d;
  double min_x, min_y, max_x, max_y;
  double x, y;
  double kx, ky;
      if (wcmNX.Getnum() < 2) return TRUE;
      if (wcmNY.Getnum() < 2) return TRUE;
      if (!wcmpdfii) return TRUE;
    wcmxStep = wcmxLength / (wcmNX.Getnum()-1); //??? Change!
    wcmyStep = wcmyLength / (wcmNY.Getnum()-1);
    d=sqrt(wcmxLength*wcmxLength+wcmyLength*wcmyLength);
  PixLenX = wcmr.right  - wcmr.left;
  PixLenY = wcmr.bottom - wcmr.top;
  if ( wcmfMin >= wcmfMax ) wcmMinMax();
  if ( wcmfMin >= wcmfMax )
    wcmScale1 = 1.0;
  else
    wcmScale1 = d*wcmScale/(wcmfMax-wcmfMin);
  m1 = 0; // m1 = wcmNY.Getna()-wcmNY.Getna();
  n1 = 0; // n1 = wcmNX.Getna()-wcmNX.Getna();
  wcmTransform(m1, n1, wcmpdfii(wcmNY.Getna(),wcmNX.Getna())-wcmfMin, &min_x, &min_y);
  max_x = min_x;
  max_y = min_y;
  for (m = wcmNY.Getna(); m < wcmNY.Getnbp1(); m++)
    for (n = wcmNX.Getna(); n < wcmNX.Getnbp1(); n++)
    {
      m1 = m - wcmNY.Getna();
      n1 = n - wcmNX.Getna();
      wcmTransform(m1, n1, wcmpdfii(m,n)-wcmfMin, &x, &y);
      if (x < min_x) min_x = x; else if (x > max_x) max_x = x;
      if (y < min_y) min_y = y; else if (y > max_y) max_y = y;
    }
  kx = PixLenX / (max_x - min_x);
  ky = PixLenY / (max_y - min_y);
  switch (wcmMode)
  {
  case wcmLINE:
  for (m = wcmNY.Getnb(); m >= wcmNY.Getna(); m--)
  {
    m1 = m - wcmNY.Getna();
    wcmTransform(m1, 0, wcmpdfii(m,wcmNX.Getna())-wcmfMin, &x, &y);
    mx = kx * (x - min_x);
    my = ky * (y - min_y);
    MoveToEx(pws->WShDC, wcmr.left + mx, wcmr.bottom - my, NULL);
    for (n = wcmNX.Getna()+1; n < wcmNX.Getnbp1(); n++)
    {
      n1 = n - wcmNX.Getna();
      wcmTransform(m1, n1, wcmpdfii(m,n)-wcmfMin, &x, &y);
      mx = kx * (x - min_x);
      my = ky * (y - min_y);
      LineTo(pws->WShDC, wcmr.left + mx, wcmr.bottom - my);
    }
  }

  for (n = wcmNX.Getnb(); n >= wcmNX.Getna(); n--)
  {
    n1 = n - wcmNX.Getna();
    wcmTransform(0, n1, wcmpdfii(wcmNY.Getna(),n)-wcmfMin, &x, &y);
    mx = kx * (x - min_x);
    my = ky * (y - min_y);
    MoveToEx(pws->WShDC, wcmr.left + mx, wcmr.bottom - my, NULL);
    for (m = wcmNY.Getna()+1; m < wcmNY.Getnbp1(); m++)
    {
      m1 = m - wcmNY.Getna();
      wcmTransform(m1, n1, wcmpdfii(m,n)-wcmfMin, &x, &y);
      mx = kx * (x - min_x);
      my = ky * (y - min_y);
      LineTo(pws->WShDC, wcmr.left + mx, wcmr.bottom - my);
    }
  }
  break;
  case wcmBRUSH:
  {
  POINT pnt[4];
  HBRUSH obr,bbr,wbr;
  int norm_y;
    bbr = CreateSolidBrush(wcmColor1);
    wbr = CreateSolidBrush(wcmColor2);
    obr = (HBRUSH)SelectObject(pws->WShDC, wbr);
    for (m = wcmNY.Getnb(); m > wcmNY.Getna(); m--)
    {
      m1 = m - wcmNY.Getna();
      n1 = wcmNX.Getnb() - wcmNX.Getna();
      wcmTransform(m1, n1, wcmpdfii(m,wcmNX.Getnb())-wcmfMin, &x, &y);
      mx = kx * (x - min_x);
      my = ky * (y - min_y);
      pnt[1].x = wcmr.left   + mx;
      pnt[1].y = wcmr.bottom - my;
      wcmTransform(m1-1, n1, wcmpdfii(m-1,wcmNX.Getnb())-wcmfMin, &x, &y);
      mx = kx * (x - min_x);
      my = ky * (y - min_y);
      pnt[2].x = wcmr.left   + mx;
      pnt[2].y = wcmr.bottom - my;
      for (n = wcmNX.Getnb()-1; n >= wcmNX.Getna(); n--)
      {
        n1 = n - wcmNX.Getna();
        pnt[0].x = pnt[1].x;
        pnt[0].y = pnt[1].y;
        pnt[3].x = pnt[2].x;
        pnt[3].y = pnt[2].y;
        wcmTransform(m1, n1, wcmpdfii(m,n)-wcmfMin, &x, &y);
        mx = kx * (x - min_x);
        my = ky * (y - min_y);
        pnt[1].x = wcmr.left   + mx;
        pnt[1].y = wcmr.bottom - my;
        wcmTransform(m1-1, n1, wcmpdfii(m-1,n)-wcmfMin, &x, &y);
        mx = kx * (x - min_x);
        my = ky * (y - min_y);
        pnt[2].x = wcmr.left   + mx;
        pnt[2].y = wcmr.bottom - my;
        norm_y = (pnt[1].x - pnt[0].x)*(pnt[2].y - pnt[1].y) -
                 (pnt[2].x - pnt[1].x)*(pnt[1].y - pnt[0].y);
        if (norm_y < 0)
          SelectObject(pws->WShDC, wbr);
        else
          SelectObject(pws->WShDC, bbr);
       Polygon(pws->WShDC, pnt, 4);
     }
   }
   SelectObject(pws->WShDC, obr);
   DeleteObject(bbr);
   DeleteObject(wbr);
  }
  break;
  }
  return FALSE;
}

//####################################################################\\
void WINCOLORMAP::PrepareColor12(int& Red1, int& Green1, int& Blue1,
                                 int& Red2, int& Green2, int& Blue2 )
{
    Red1   = GetRValue ( wcmColor1 );
    Green1 = GetGValue ( wcmColor1 );
    Blue1  = GetBValue ( wcmColor1 );
    Red2   = GetRValue ( wcmColor2 );
    Green2 = GetGValue ( wcmColor2 );
    Blue2  = GetBValue ( wcmColor2 );
}
//####################################################################\\
void WINCOLORMAP::SetNumOfColors(int ne)
{
  if ( ne < 1 ) ne = 1;
  if ( wcmRGB != NULL ) delete[] wcmRGB;
  wcmNumOfColors = ne;
  wcmRGB = new COLORREF[wcmNumOfColors];
  PrepareRGB();
}
//####################################################################\\
void WINCOLORMAP::PrepareRGB(void)
{
  int i, Red, Green, Blue;
  int Red1,Red2,Green1,Green2,Blue1,Blue2;
  PrepareColor12(Red1, Green1, Blue1, Red2, Green2, Blue2 );
  if (!wcmRGB) return;
  wcmRGB[0]=RGB(Red1,Green1,Blue1);
  if ( wcmNumOfColors > 1) for (i = 1; i < wcmNumOfColors; i++)
  {
    Red   = Red1  +(Red2  -Red1  )*i/(wcmNumOfColors-1);
    Green = Green1+(Green2-Green1)*i/(wcmNumOfColors-1);
    Blue  = Blue1 +(Blue2 -Blue1 )*i/(wcmNumOfColors-1);
    wcmRGB[i]=RGB(Red,Green,Blue);
  }
}

//####################################################################\\
void WINCOLORMAP::AllocCellInterpolant()
{
  int i, j, k, p;
  int pp = PixCellX * PixCellY;
  lWeightMax = 16000;
  lValueMax = 16000L;
  if ( pp > 4000 ){ pp = 1; wcmInterpFlag = FALSE;} else  wcmInterpFlag = TRUE;
  px1y1 = new int[pp];
  px2y1 = new int[pp];
  px2y2 = new int[pp];
  px1y2 = new int[pp];
  if ( !wcmInterpFlag ) return;
  if ( PixCellX == 1 && PixCellY == 1)
  {
     px1y1[0] = lWeightMax;
     px2y1[0] = 0;
     px2y2[0] = 0;
     px1y2[0] = 0;
     return;
  }
  if ( PixCellX == 1 )
  {
    p = 0;
    k = PixCellY-1;
    lWeightMax = (lWeightMax/k)*k;
    for (j=0; j < PixCellY; j++)
    {
      px1y1[p] = (PixCellY-j-1) * lWeightMax/k;
      px2y1[p] = 0;
      px2y2[p] = 0;
      px1y2[p] = j * lWeightMax/k;
      p++;
    }
    return;
  }
  if ( PixCellY == 1 )
  {
    p = 0;
    k = PixCellX-1;
    lWeightMax = (lWeightMax/k)*k;
    for (i=0; i < PixCellX; i++)
    {
      px1y1[p] = (PixCellX-i-1) * lWeightMax/k;
      px2y1[p] = i * lWeightMax/k;
      px2y2[p] = 0;
      px1y2[p] = 0;
      p++;
    }
    return;
  }
  p = 0;
  k = (PixCellX-1) * (PixCellY-1);
  lWeightMax = (lWeightMax/k)*k;
  for (i=0; i < PixCellX; i++)
  for (j=0; j < PixCellY; j++)
  {
//    p = i + j*PixCellX;
    px1y1[p] = ((PixCellX-i-1)*(PixCellY-j-1))*lWeightMax/k;
    px2y1[p] = (i             *(PixCellY-j-1))*lWeightMax/k;
    px2y2[p] = (i             *j             )*lWeightMax/k;
    px1y2[p] = ((PixCellX-i-1)*j             )*lWeightMax/k;
    p++;
  }
  return;
}

//####################################################################\\
void WINCOLORMAP::DeAllocCellInterpolant(void)
{
  if (!px1y1) {delete[] px1y1; px1y1 = NULL;}
  if (!px1y2) {delete[] px1y2; px1y2 = NULL;}
  if (!px2y1) {delete[] px2y1; px2y1 = NULL;}
  if (!px2y2) {delete[] px2y2; px2y2 = NULL;}
}

//####################################################################\\
void WINCOLORMAP::PaintCellBitmap(){
  static int p,i,j,k, curval, p11, p12, p21, p22;
  if (wcmInterpFlag){
    p = 0;
    for (i=0; i < PixCellX; i++) for (j=0; j < PixCellY; j++){
//      p = i + j*PixCellX;
      curval = ((
        (px1y2[p]*cx1y2 + px2y2[p]*cx2y2 + px2y1[p]*cx2y1 + px1y1[p]*cx1y1)
        /lWeightMax)*(wcmNumOfColors+1))/lValueMax;
      if (curval < 0) curval = 0;
      else if (curval >= wcmNumOfColors) curval = wcmNumOfColors-1;
      SetPixel(pws->WShDC, x1+i, y1+j, wcmRGB[curval]);
      p++;
    }
  }
  else{
    k = (PixCellX-1)*(PixCellY-1);
    for (i=0; i < PixCellX; i++) for (j=0; j < PixCellY; j++){
      p11 = (PixCellX-i-1)*(PixCellY-j-1);
      p21 = i             *(PixCellY-j-1);
      p22 = i             *j             ;
      p12 = (PixCellX-i-1)*j             ;
      curval = (((p12*cx1y2 + p22*cx2y2 + p21*cx2y1 + p11*cx1y1)/k)
       * (wcmNumOfColors+1))/lValueMax;
      if (curval < 0) curval = 0;
      else if (curval >= wcmNumOfColors) curval = wcmNumOfColors-1;
      SetPixel(pws->WShDC, x1+i, y1+j, wcmRGB[curval]);
    }
  }
}

//####################################################################\\
int WINCOLORMAP::PaintColorMap2Da()
{
  int m, n;
  int PixLenX, PixLenY;
  double amp;
  int x2m, y2m, y1t, y2t;
  if (wcmNX.Getnum() < 2) return TRUE;
  if (wcmNY.Getnum() < 2) return TRUE;
  if (!wcmpdfii) return TRUE;
//???
  if (wcmfMax==wcmfMin)
    amp = 0;
  else
    amp = double(lValueMax)/(wcmfMax - wcmfMin);
  PixLenX = wcmr.right  - wcmr.left;
  PixLenY = wcmr.bottom - wcmr.top;
  PixCellX = (PixLenX-1)/(wcmNX.Getnum()-1) + 1;
  PixCellY = (PixLenY-1)/(wcmNY.Getnum()-1) + 1;
  AllocCellInterpolant();
  x2 = wcmr.left + 1;
  for (n = wcmNX.Getna(); n < wcmNX.Getnb(); n++)
  {
    x1 = x2 - 1;
    x2 += PixCellX-1;
    x2m = wcmr.left + ((n-wcmNX.Getna()+1)*PixLenX)/(wcmNX.Getnum()-1);
    if ( x2 < x2m ){x1++; x2++;}
    y2t = wcmr.top + 1;
    for (m = wcmNY.Getna(); m < wcmNY.Getnb(); m++)
    {
      y1t = y2t - 1;
      y2t += PixCellY-1;
      y2m = wcmr.top + ((m-wcmNY.Getna()+1)*PixLenY)/(wcmNY.Getnum()-1);
      if ( y2t < y2m ){y1t++; y2t++;}
      y1 = wcmr.bottom - (y2t - wcmr.top);
      y2 = wcmr.bottom - (y1t - wcmr.top);
      cx1y2 = amp*( wcmpdfii(m,n)     - wcmfMin);
      cx2y2 = amp*( wcmpdfii(m,n+1)   - wcmfMin);
      cx2y1 = amp*( wcmpdfii(m+1,n+1) - wcmfMin);
      cx1y1 = amp*( wcmpdfii(m+1,n)   - wcmfMin);
      PaintCellBitmap();
//      SetPixel(pws->WShDC, x1, y1, RGB(255,0,0));
    }
  }
  DeAllocCellInterpolant();
    sprintf(stdbuf1, "%8.3lg", wcmfMin);
    TextOut(pws->WShDC, wcmr.left+3, wcmr.top+3,  stdbuf1, strlen(stdbuf1));
    sprintf(stdbuf1, "%8.3lg", wcmfMax);
    TextOut(pws->WShDC, wcmr.left+3, wcmr.top+18, stdbuf1, strlen(stdbuf1));
  return FALSE;
}

//--------------------------------------------\\
//  ind==0: no roots;                         \\
//  ind==1: one root==xcurr;                  \\
//  ind==2: two roots, xprev;xnext;           \\
//--------------------------------------------\\
//####################################################################\\
int support1 (int uprev, int xprev, int unext, int xnext, int * xcurr)
{
  static int ind;
  ind = 0;
  if (uprev == 0)
  {
    if (unext == 0){ ind = 2; * xcurr = (xprev+xnext)/2;}
    else           { ind = 1; * xcurr = xprev;}
  }
  else
  {
    if (unext == 0){ ind = 1; * xcurr = xnext;}
    else
    {
      if ((uprev < 0 && unext < 0) || (uprev > 0 && unext > 0))
                   { ind = 0; * xcurr = (xprev+xnext)/2;}
      else         { ind = 1; * xcurr = (unext*xprev-uprev*xnext)/(unext-uprev);}
    }
  }
  return ind;
}
//####################################################################\\
int support1 (int uprev, double xprev, int unext, double xnext, double * xcurr)
{
  static int ind;
  ind = 0;
  if (uprev == 0)
  {
    if (unext == 0){ ind = 2; * xcurr = (xprev+xnext)/2;}
    else           { ind = 1; * xcurr = xprev;}
  }
  else
  {
    if (unext == 0){ ind = 1; * xcurr = xnext;}
    else
    {
      if ((uprev < 0 && unext < 0) || (uprev > 0 && unext > 0))
                   { ind = 0; * xcurr = (xprev+xnext)/2;}
      else         { ind = 1; * xcurr = (unext*xprev-uprev*xnext)/(unext-uprev);}
    }
  }
  return ind;
}

//####################################################################\\
void WINCOLORMAP::gline_w ( int x1, int y1, int x2, int y2 )
{
  MoveToEx (pws->WShDC, x1, y1, NULL);
  LineTo   (pws->WShDC, x2, y2);
}

//####################################################################\\
int WINCOLORMAP::PaintIzolin2Da(PDFUNCLL pfll, int ndy, int ndx, double dfMin, double dfMax)
{
  SetNY(ndy);
  SetNX(ndx);
  SetfMin(dfMin);
  SetfMax(dfMax);
  SetpFunc(pfll);
  return PaintIzolin2Da();
}

//####################################################################\\
int WINCOLORMAP::PaintIzolin2Da()
{
    int iLevel;
    int ind12, ind34, ind13, ind24;
    int aCurLevel;
    int x12, x34, y13, y24;
    int ux1y1, ux2y1, ux1y2, ux2y2;
    int u0x1y1, u0x2y1, u0x1y2, u0x2y2;
  int PixLenX,PixLenY;
  double amp, dfMin, dfMax;
  int x2m, y2m;
  int m, n, n1, m1;
  dfMin = GetfMin();
  dfMax = GetfMax();
  lValueMax = 16000L;
  if ( dfMax < dfMin ){ amp = dfMax; dfMax = dfMin; dfMin = amp;}
  if ( dfMax - dfMin < 1.0e-19 ) dfMax = dfMin + 1.0e-19;
  if ( dfMax - dfMin < 1.0e-19 ) dfMax = dfMin + fabs(dfMin) * 1.0e-9;
  amp = lValueMax/(dfMax - dfMin);
  HPEN* hhPen = new HPEN[wcmNumOfColors];
  for ( iLevel = 0; iLevel < wcmNumOfColors; iLevel++ )
     hhPen[iLevel] = CreatePen ( PS_SOLID, 1, wcmRGB[iLevel]);
  HPEN hPenOld = SelectPen(pws->WShDC,hhPen[0]);
  PixLenX = wcmr.right - wcmr.left;
  PixLenY = wcmr.bottom - wcmr.top;
  PixCellX = (PixLenX-1)/(wcmNX.Getnum()-1)+1;
  PixCellY = (PixLenY-1)/(wcmNY.Getnum()-1)+1;
  x2 = wcmr.left+1;
  for (n = 0; n < wcmNX.Getnum()-1; n++)
  {
    n1 = n + wcmNX.Getna();
    x1 = x2-1;
    x2 += (PixCellX-1);
    x2m = wcmr.left + int(n+1)*int(PixLenX)/int(wcmNX.Getnum()-1);
    if ( x2 < x2m ){x2++;}
      ux1y2 = amp * (wcmpdfii(0,n1)   - dfMin);
      ux2y2 = amp * (wcmpdfii(0,n1+1) - dfMin);
        if ( ux1y2 < 0      ) ux1y2 = 0;
        if ( ux1y2 >= lValueMax ) ux1y2 = lValueMax-1;
        if ( ux2y2 < 0      ) ux2y2 = 0;
        if ( ux2y2 >= lValueMax ) ux2y2 = lValueMax-1;
    y2 = wcmr.top+1;
    for (m = 0; m < wcmNY.Getnum()-1; m++)
    {
      m1 = m + wcmNY.Getna();
      y1 = y2-1;
      y2 += (PixCellY-1);
      y2m = wcmr.top + int(m+1)*int(PixLenY)/int(wcmNY.Getnum()-1);
      if ( y2 < y2m ){y2++;}
        ux1y1 = ux1y2;
        ux2y1 = ux2y2;
        ux1y2 = amp * (wcmpdfii(m1+1,n1)   - dfMin);
        ux2y2 = amp * (wcmpdfii(m1+1,n1+1) - dfMin);
          if ( ux1y2 < 0      ) ux1y2 = 0;
          if ( ux1y2 >= lValueMax ) ux1y2 = lValueMax-1;
          if ( ux2y2 < 0      ) ux2y2 = 0;
          if ( ux2y2 >= lValueMax ) ux2y2 = lValueMax-1;
      for ( iLevel = 0; iLevel < wcmNumOfColors; iLevel++ )
      {
        aCurLevel = lValueMax * iLevel / (wcmNumOfColors-1);
//             if ( Regime == IZOREGIONS )
//               aCurLevel = 0.5 * ( aCurLevel + fVal ( iLevel +1 ) );
        u0x1y1 = ux1y1 - aCurLevel;
        u0x2y1 = ux2y1 - aCurLevel;
        u0x1y2 = ux1y2 - aCurLevel;
        u0x2y2 = ux2y2 - aCurLevel;
        if ((u0x1y1 < 0)&&(u0x2y1 < 0)&&(u0x1y2 < 0)&&(u0x2y2 < 0)) continue;
        if ((u0x1y1 > 0)&&(u0x2y1 > 0)&&(u0x1y2 > 0)&&(u0x2y2 > 0)) continue;
        ind12 = support1(u0x1y1, x1, u0x2y1, x2, &x12);
        ind34 = support1(u0x1y2, x1, u0x2y2, x2, &x34);
        ind13 = support1(u0x1y1, y1, u0x1y2, y2, &y13);
        ind24 = support1(u0x2y1, y1, u0x2y2, y2, &y24);
        SelectPen (pws->WShDC, hhPen[iLevel]);
//        if ( ind12 == 2 ) gline_w ( x1, y1, x2, y1 );
//        if ( ind34 == 2 ) gline_w ( x1, y2, x2, y2 );
//        if ( ind13 == 2 ) gline_w ( x1, y1, x1, y2 );
//        if ( ind24 == 2 ) gline_w ( x2, y1, x2, y2 );
        if ( ind12 == 1 )
        {
          if ( ind24 == 1 ) gline_w ( x12, y1, x2, y24 );
          if ( ind34 == 1 ) gline_w ( x12, y1, x34, y2  );
          if ( ind13 == 1 ) gline_w ( x12, y1, x1, y13 );
        }
        if ( ind24 == 1 )
        {
          if ( ind34 == 1 ) gline_w ( x2, y24, x34, y2  );
          if ( ind13 == 1 ) gline_w ( x2, y24, x1, y13 );
        }
        if ( ind34 == 1 )
        {
          if ( ind13 == 1 ) gline_w ( x34, y2, x1, y13 );
        }
      }
//      SetPixel(pws->WShDC, x1, y1, RGB(255,0,0));
    }
  }
  SelectPen(pws->WShDC, hPenOld);
  for ( iLevel = 0; iLevel < wcmNumOfColors; iLevel++ )
     DeletePen(hhPen[iLevel]);
  delete[] hhPen;
  return TRUE;
}

//####################################################################\\
int WINCOLORMAP::ftx(double dx)
{
  return (int(((1.0+dx/rhomax)*0.5)*PixLenX)+PixMinX);
}
//####################################################################\\
int WINCOLORMAP::fty(double dy)
{
  return (int(((1.0+dy/rhomax)*0.5)*PixLenY)+PixMinY);
}
//####################################################################\\
int WINCOLORMAP::PaintIzolin2Dc(
  PDFUNCLL pfllNew,
  double   dfMinNew, double dfMaxNew,
  PDFUNCLL pfllOld,
  double   dfMinOld, double dfMaxOld,
  PDFUNCL  frhoPoints,
  PDFUNCL  fphiPoints,
  int Lrho, int Mrho, int Lphi, int Mphi )
{
  int iLevel;
  int ind12, ind34, ind13, ind24;
  int aCurLevel;
//  int ux1y1, ux2y1, ux1y2, ux2y2, u0x1y1, u0x2y1, u0x1y2, u0x2y2;
  double amp;
  double ampOld;
//  int x2m, y2m;

  double phi, rhoa, rhob, phia, phib;
  double cphia, sphia, cphib, sphib;
  double rho12, rho34, phi13, phi24;
  int x12, y12, x34, y34, x13, y13, x24, y24;
  int u1, u2, u3, u4, u1m, u2m, u3m, u4m, u01, u02, u03, u04;
    int LrhoP1, LphiP1;
    int jrho, jrhoM1, iphi;
    int nRhoPoints, nPhiPoints;

    LrhoP1 = Lrho + 1;
    LphiP1 = Lphi + 1;
  nRhoPoints = Mrho+1-Lrho;
  nPhiPoints = Mphi+1-Lphi;

  DOUBLE1DM cosphi(nPhiPoints);
  DOUBLE1DM sinphi(nPhiPoints);
  for ( iphi = Lphi; iphi <= Mphi; iphi++ )
  {
	 phi = fphiPoints(iphi);
	 cosphi[iphi-Lphi] = cos(phi);
	 sinphi[iphi-Lphi] = sin(phi);
  }
  dminmax(&rhomin, &rhomax, frhoPoints, nRhoPoints, Lrho);
  dminmax(&phimin, &phimax, fphiPoints, nPhiPoints, Lphi);

  lValueMax = 16000L;
  if ( dfMaxNew < dfMinNew ){ amp = dfMaxNew; dfMaxNew = dfMinNew; dfMinNew = amp;}
  if ( dfMaxNew - dfMinNew < 1.0e-19 ) dfMaxNew = dfMinNew + 1.0e-19;
  if ( dfMaxNew - dfMinNew < 1.0e-19 ) dfMaxNew = dfMinNew + fabs(dfMinNew) * 1.0e-9;
  amp = lValueMax/(dfMaxNew - dfMinNew);
  if ( dfMaxOld < dfMinOld ){ amp = dfMaxOld; dfMaxOld = dfMinOld; dfMinOld = amp;}
  if ( dfMaxOld - dfMinOld < 1.0e-19 ) dfMaxOld = dfMinOld + 1.0e-19;
  if ( dfMaxOld - dfMinOld < 1.0e-19 ) dfMaxOld = dfMinOld + fabs(dfMinOld) * 1.0e-9;
  ampOld = lValueMax/(dfMaxOld - dfMinOld);

  HPEN hPenBack = CreatePen( PS_SOLID, 1, wcmColorBack);
  HPEN hPenOld = SelectPen(pws->WShDC,hPenBack);
  HPEN* hhPen = new HPEN[wcmNumOfColors];
  for ( iLevel = 0; iLevel < wcmNumOfColors; iLevel++ )
     hhPen[iLevel] = CreatePen ( PS_SOLID, 1, wcmRGB[iLevel]);

  PixMinX = wcmr.left;
  PixMinY = wcmr.top;
  PixLenX = wcmr.right - wcmr.left;
  PixLenY = wcmr.bottom - wcmr.top;

  rhob = frhoPoints ( Lrho );
  for ( jrho = LrhoP1; jrho <= Mrho; jrho++ )
  {
    jrhoM1 = jrho - 1;
    rhoa = rhob;
    rhob = frhoPoints(jrho);
    phib = fphiPoints(Lphi);
    cphib = cosphi[0];
    sphib = sinphi[0];
    u3  = amp * ( pfllNew(jrhoM1, Lphi) - dfMinNew);
    u4  = amp * ( pfllNew(jrho, Lphi) - dfMinNew);
    u3m = ampOld * ( pfllOld(jrhoM1, Lphi) - dfMinOld);
    u4m = ampOld * ( pfllOld(jrho, Lphi) - dfMinOld);
    for ( iphi = LphiP1; iphi <= Mphi; iphi++ )
    {
      phia = phib;
      cphia = cphib;
      sphia = sphib;
	   phib = fphiPoints ( iphi );
	   cphib = cosphi[iphi-Lphi];
	   sphib = sinphi[iphi-Lphi];
	   u1 = u3;
	   u2 = u4;
	   u3 = amp * (pfllNew(jrhoM1, iphi) - dfMinNew);
	   u4 = amp * (pfllNew(jrho, iphi) - dfMinNew);
	   u1m = u3m;
	   u2m = u4m;
	   u3m = ampOld * (pfllOld(jrhoM1, iphi) - dfMinOld);
	   u4m = ampOld * (pfllOld(jrho, iphi) - dfMinOld);
          SelectPen (pws->WShDC, hPenBack);
//          SelectPen (wcmhdc, hPenGreen);
      for ( iLevel = 0; iLevel < wcmNumOfColors; iLevel++ )
	   {
        aCurLevel = (lValueMax * iLevel)/(wcmNumOfColors-1);
	     u01 = u1m - aCurLevel;
	     u02 = u2m - aCurLevel;
	     u03 = u3m - aCurLevel;
	     u04 = u4m - aCurLevel;
	     if ((u01<0)&&(u02<0)&&(u03<0)&&(u04<0)) continue;
	     if ((u01>0)&&(u02>0)&&(u03>0)&&(u04>0)) continue;
	     ind12 = support1 ( u01, rhoa, u02, rhob, &rho12 );
	     ind34 = support1 ( u03, rhoa, u04, rhob, &rho34 );
	     ind13 = support1 ( u01, phia, u03, phib, &phi13 );
	     ind24 = support1 ( u02, phia, u04, phib, &phi24 );
	     x1 =  ftx(rhoa * cphia);
	     y1 =  fty(rhoa * sphia);
	     x2 =  ftx(rhob * cphia);
	     y2 =  fty(rhob * sphia);
	     x4 =  ftx(rhob * cphib);
	     y4 =  fty(rhob * sphib);
	     x3 =  ftx(rhoa * cphib);
	     y3 =  fty(rhoa * sphib);
	     x12 = ftx(rho12 * cphia);
	     y12 = fty(rho12 * sphia);
	     x34 = ftx(rho34 * cphib);
	     y34 = fty(rho34 * sphib);
	     x13 = ftx(rhoa * cos(phi13));
	     y13 = fty(rhoa * sin(phi13));
	     x24 = ftx(rhob * cos(phi24));
	     y24 = fty(rhob * sin(phi24));
//	     if ( ind12 == 2 ) gline_w ( x1, y1, x2, y2 );
//	     if ( ind34 == 2 ) gline_w ( x3, y3, x4, y4 );
//	     if ( ind13 == 2 ) gline_w ( x1, y1, x3, y3 );
//	     if ( ind24 == 2 ) gline_w ( x2, y2, x4, y4 );
	     if ( ind12 == 1 )
	     {
          if ( ind24 == 1 ) gline_w ( x12, y12, x24, y24 );
          if ( ind34 == 1 ) gline_w ( x12, y12, x34, y34 );
          if ( ind13 == 1 ) gline_w ( x12, y12, x13, y13 );
        }
	     if ( ind24 == 1 )
	     {
		    if ( ind34 == 1 ) gline_w ( x24, y24, x34, y34 );
          if ( ind13 == 1 ) gline_w ( x24, y24, x13, y13 );
	     }
	     if ( ind34 == 1 )
	     {
		    if ( ind13 == 1 ) gline_w ( x34, y34, x13, y13 );
	     }
	   }
      for ( iLevel = 0; iLevel < wcmNumOfColors; iLevel++ )
	   {
        aCurLevel = (lValueMax * iLevel)/(wcmNumOfColors-1);
	     u01 = u1 - aCurLevel;
	     u02 = u2 - aCurLevel;
	     u03 = u3 - aCurLevel;
	     u04 = u4 - aCurLevel;
	     if ((u01<0)&&(u02<0)&&(u03<0)&&(u04<0)) continue;
	     if ((u01>0)&&(u02>0)&&(u03>0)&&(u04>0)) continue;
        SelectPen (pws->WShDC, hhPen[iLevel]);

	     ind12 = support1 ( u01, rhoa, u02, rhob, &rho12 );
	     ind34 = support1 ( u03, rhoa, u04, rhob, &rho34 );
	     ind13 = support1 ( u01, phia, u03, phib, &phi13 );
	     ind24 = support1 ( u02, phia, u04, phib, &phi24 );
	     x1  = ftx(rhoa * cphia);
	     y1  = fty(rhoa * sphia);
	     x2  = ftx(rhob * cphia);
	     y2  = fty(rhob * sphia);
	     x4  = ftx(rhob * cphib);
	     y4  = fty(rhob * sphib);
	     x3  = ftx(rhoa * cphib);
	     y3  = fty(rhoa * sphib);
	     x12 = ftx(rho12 * cphia);
	     y12 = fty(rho12 * sphia);
	     x34 = ftx(rho34 * cphib);
	     y34 = fty(rho34 * sphib);
	     x13 = ftx(rhoa * cos(phi13));
	     y13 = fty(rhoa * sin(phi13));
	     x24 = ftx(rhob * cos(phi24));
	     y24 = fty(rhob * sin(phi24));
/*
	     if ( 0 )
        {
        SelectPen (wcmhdc, hPenGreen);
          gline_w ( x1, y1, x2, y2 );
	       gline_w ( x3, y3, x4, y4 );
	       gline_w ( x1, y1, x3, y3 );
	       gline_w ( x2, y2, x4, y4 );
        SelectPen (wcmhdc, hPenBlue);
          gline_w ( x1, y1, x4, y4 );
	       gline_w ( x3, y3, x2, y2 );
        }
*/
//	     if ( ind12 == 2 ) gline_w ( x1, y1, x2, y2 );
//	     if ( ind34 == 2 ) gline_w ( x3, y3, x4, y4 );
//	     if ( ind13 == 2 ) gline_w ( x1, y1, x3, y3 );
//	     if ( ind24 == 2 ) gline_w ( x2, y2, x4, y4 );

	     if ( ind12 == 1 )
	     {
          if ( ind24 == 1 ) gline_w ( x12, y12, x24, y24 );
          if ( ind34 == 1 ) gline_w ( x12, y12, x34, y34 );
          if ( ind13 == 1 ) gline_w ( x12, y12, x13, y13 );
	     }
	     if ( ind24 == 1 )
	     {
          if ( ind34 == 1 ) gline_w ( x24, y24, x34, y34 );
          if ( ind13 == 1 ) gline_w ( x24, y24, x13, y13 );
	     }
	     if ( ind34 == 1 )
	     {
		    if ( ind13 == 1 ) gline_w ( x34, y34, x13, y13 );
	     }
	   }
    }
  }

/*

  x2 = wcmr.left+1;
  for (i = 0; i < ndpx-1; i++)
  {
    x1 = x2-1;
    x2 += (PixCellX-1);
    x2m = wcmr.left + int(i+1)*int(PixLenX)/int(ndpx-1);
    if ( x2 < x2m ){x2++;}
      ux1y2 = amp * (pfll(0,i)   - dfMin);
      ux2y2 = amp * (pfll(0,i+1) - dfMin);
        if ( ux1y2 < 0      ) ux1y2 = 0;
        if ( ux1y2 >= lValueMax ) ux1y2 = lValueMax-1;
        if ( ux2y2 < 0      ) ux2y2 = 0;
        if ( ux2y2 >= lValueMax ) ux2y2 = lValueMax-1;
    y2 = wcmr.top+1;
    for (j = 0; j < ndpy-1; j++)
    {
      y1 = y2-1;
      y2 += (PixCellY-1);
      y2m = wcmr.top + int(j+1)*int(PixLenY)/int(ndpy-1);
      if ( y2 < y2m ){y2++;}
        ux1y1 = ux1y2;
        ux2y1 = ux2y2;
        ux1y2 = amp * (pfll(j+1,i)   - dfMin);
        ux2y2 = amp * (pfll(j+1,i+1) - dfMin);
          if ( ux1y2 < 0      ) ux1y2 = 0;
          if ( ux1y2 >= lValueMax ) ux1y2 = lValueMax-1;
          if ( ux2y2 < 0      ) ux2y2 = 0;
          if ( ux2y2 >= lValueMax ) ux2y2 = lValueMax-1;
        }
      }
    }
  }

*/
  SelectPen(pws->WShDC, hPenOld);
  for ( iLevel = 0; iLevel < wcmNumOfColors; iLevel++ )
     DeletePen(hhPen[iLevel]);
  delete[] hhPen;
  DeletePen(hPenBack);
  return TRUE;
}

//####################################################################\\
int WINCOLORMAP::PaintColorMap2Dc(
  PDFUNCLL pfll, PDFUNCL  frhoPoints, PDFUNCL  fphiPoints,
  int Lrho, int Mrho, int Lphi, int Mphi )
{
  int iLevel;
  double amp;
  double phi, rhoa, rhob;
  double cphia, sphia, cphib, sphib;
  int u1, u2, u3, u4;
  int LrhoP1, LphiP1;
  int jrho, jrhoM1, iphi;
  int nRhoPoints, nPhiPoints;
  double dfMinTemp, dfMaxTemp;
  POINT Points[4];
    LrhoP1 = Lrho + 1;
    LphiP1 = Lphi + 1;
  nRhoPoints = Mrho+1-Lrho;
  nPhiPoints = Mphi+1-Lphi;
    DOUBLE1DM cosphi(nPhiPoints);
    DOUBLE1DM sinphi(nPhiPoints);
  for ( iphi = Lphi; iphi <= Mphi; iphi++ )
  {
	 phi = fphiPoints(iphi);
	 cosphi[iphi-Lphi] = cos(phi);
	 sinphi[iphi-Lphi] = sin(phi);
  }
  dminmax(&rhomin, &rhomax, frhoPoints, nRhoPoints, Lrho);
  dminmax(&phimin, &phimax, fphiPoints, nPhiPoints, Lphi);

  lValueMax = 16000L;
  dfMinTemp = wcmfMin;
  dfMaxTemp = wcmfMax;
  if ( dfMaxTemp < dfMinTemp ){ amp = dfMaxTemp; dfMaxTemp = dfMinTemp; dfMinTemp = amp;}
  if ( dfMaxTemp - dfMinTemp < 1.0e-19 ) dfMaxTemp = dfMinTemp + 1.0e-19;
  if ( dfMaxTemp - dfMinTemp < 1.0e-19 ) dfMaxTemp = dfMinTemp + fabs(dfMinTemp) * 1.0e-9;
  amp = lValueMax/(dfMaxTemp - dfMinTemp);

  HPEN hPenBack = CreatePen( PS_SOLID, 1, wcmColorBack);
  HPEN hPenOld = SelectPen(pws->WShDC, hPenBack);
  HPEN* hhPen = new HPEN[wcmNumOfColors];
  for ( iLevel = 0; iLevel < wcmNumOfColors; iLevel++ )
    hhPen[iLevel] = CreatePen ( PS_SOLID, 1, wcmRGB[iLevel]);
  HBRUSH hbrBack = CreateSolidBrush (wcmColorBack);
  HBRUSH hbrOld = (HBRUSH)SelectObject(pws->WShDC, hbrBack);
  HBRUSH* hhbr = new HBRUSH[wcmNumOfColors];
  for ( iLevel = 0; iLevel < wcmNumOfColors; iLevel++ )
    hhbr[iLevel] = CreateSolidBrush (wcmRGB[iLevel]);
  PixMinX = wcmr.left;
  PixMinY = wcmr.top;
  PixLenX = wcmr.right - wcmr.left;
  PixLenY = wcmr.bottom - wcmr.top;
    rhob = frhoPoints ( Lrho );
  for ( jrho = LrhoP1; jrho <= Mrho; jrho++ )
  {
    jrhoM1 = jrho - 1;
    rhoa = rhob;
    rhob = frhoPoints(jrho);
//    phib = fphiPoints(Lphi);
    cphib = cosphi[0];
    sphib = sinphi[0];
    u3  = amp * ( pfll(jrhoM1, Lphi) - dfMinTemp);
    u4  = amp * ( pfll(jrho,   Lphi) - dfMinTemp);
    for ( iphi = LphiP1; iphi <= Mphi; iphi++ )
    {
//      phia = phib;
      cphia = cphib;
      sphia = sphib;
//	   phib = fphiPoints ( iphi );
	   cphib = cosphi[iphi-Lphi];
	   sphib = sinphi[iphi-Lphi];
	   u1 = u3;
	   u2 = u4;
	   u3 = amp * (pfll(jrhoM1, iphi) - dfMinTemp);
	   u4 = amp * (pfll(jrho,   iphi) - dfMinTemp);
        SelectPen (pws->WShDC, hPenBack);
        SelectObject(pws->WShDC, hbrBack);
        iLevel = (u1+u2+u3+u4)*(wcmNumOfColors-1)/4/lValueMax;
        if ( iLevel < 0 ) iLevel = 0;
        if ( iLevel >= wcmNumOfColors ) iLevel = wcmNumOfColors-1;
        SelectPen (pws->WShDC, hhPen[iLevel]);
        SelectObject(pws->WShDC, hhbr[iLevel]);
	     x1  = ftx(rhoa * cphia);
	     y1  = fty(rhoa * sphia);
	     x2  = ftx(rhob * cphia);
	     y2  = fty(rhob * sphia);
	     x4  = ftx(rhob * cphib);
	     y4  = fty(rhob * sphib);
	     x3  = ftx(rhoa * cphib);
	     y3  = fty(rhoa * sphib);
        Points[0].x = x1;
        Points[0].y = y1;
        Points[3].x = x2;
        Points[3].y = y2;
        Points[1].x = x3;
        Points[1].y = y3;
        Points[2].x = x4;
        Points[2].y = y4;
        Polygon(pws->WShDC, Points, 4);
    }
  }
  SelectObject(pws->WShDC, hbrOld);
  for ( iLevel = 0; iLevel < wcmNumOfColors; iLevel++ )
     DeleteObject(hhbr[iLevel]);
  delete[] hhbr;
  DeleteObject(hbrBack);
  SelectPen(pws->WShDC, hPenOld);
  for ( iLevel = 0; iLevel < wcmNumOfColors; iLevel++ )
     DeletePen(hhPen[iLevel]);
  delete[] hhPen;
  DeletePen(hPenBack);
  return TRUE;
}

//####################################################################\\
//####################################################################\\
//####################################################################\\

// there is another constructor!
#pragma argsused
WINSTON::WINSTON(GWINPROC mgwp, HWND mGWhWnd, HDC mGWhDC,
   int      mGWThisIndex
,  char   * mGWClassName
,  char   * mGWWindowTitle
,  PIXRECT  mWSDefRect
,  int mState){
  GWinProc      = mgwp;
  WShWnd        = mGWhWnd;
  GWState       = GWS_CLOZED;
  WShDC         = mGWhDC;
  WSClassName   = mGWClassName;
  GWWindowTitle = mGWWindowTitle;
  GWThisIndex   = mGWThisIndex;
//???  GWWindPos     = WINDPOS(0,0,640,480);
  WSDefRect = mWSDefRect;
  WSSetDefSize(36, 24);
    WSRect.left = 0;
    WSRect.right= 640-1;
    WSRect.top = 0;
    WSRect.bottom = 480-1;
}

// there is another constructor!
#pragma argsused
WINSTON::WINSTON(GWINPROC mgwp, HWND mGWhWnd, HDC mGWhDC,
   int      mGWThisIndex
,  char   * mGWClassName
,  char   * mGWWindowTitle
,  SSEGM  ds
,  int mState
,  PIXSIZE psize){
  GWinProc      = mgwp;
  WShWnd        = mGWhWnd;
  GWState       = GWS_CLOZED;
  WShDC         = mGWhDC;
  WSClassName   = mGWClassName;
  GWWindowTitle = mGWWindowTitle;
  GWThisIndex   = mGWThisIndex;
//???  GWWindPos     = WINDPOS(0,0,640,480);
  WSDefRect = PIXRECT(ds.l, ds.t, ds.r, ds.b);
  WSSetDefSize(psize.Geth(), psize.Getv());
	 WSRect.left = 0;
	 WSRect.right= 640-1;
	 WSRect.top = 0;
	 WSRect.bottom = 480-1;
}

//####################################################################\\
//####################################################################\\
//  COMMON int nPartsStore,  nCurStep;
//  COMMON PPDOUBLE xStore, yStore, zStore;

//####################################################################\\

//####################################################################\\
PPWCM AllocWCM2D (int rows , int cols){
  PPWCM matrix; int i, j;
  matrix = new PWCM [rows];
  for (i = 0; i < rows; i++)
     matrix[i] = new WINCOLORMAP[cols];
  for (i = 0; i < rows; i++)
     for (j = 0; j < cols; j++)
        ; // matrix [i][j] = 0.0;
  membyte += rows * cols * sizeof(WINCOLORMAP);
  return matrix;
}
//####################################################################\\
void DeAllocWCM2D (PPWCM& x, int rows){
  int i;
  if (x == NULL ) return;
  for (i = 0; i < rows; i++)
    if (x[i] != NULL){
      delete[] x[i];
      x[i] = NULL;
    }
  delete[] x;
  x = NULL;
}
//####################################################################\\
void ReAllocWCM2D (PPWCM & x, int rows, int columns){
  DeAllocWCM2D(x, rows);
  x = AllocWCM2D(rows, columns);
}
//####################################################################\\
void ProcessWCM2D (PPWCM & x, int rows, int columns, int what){
  switch(what){
  case STD_ALLOC:
    x = AllocWCM2D(rows, columns);
    break;
  case STD_DEALLOC:
    DeAllocWCM2D(x, rows);
    break;
  case STD_REALLOC:
    ReAllocWCM2D(x, rows, columns);
    break;
  default:
    MessageBox(NULL, "ProcessWCM2D", "Error", MB_OK);
    break;
  }
}
//####################################################################\\
PWCM AllocWCM1D (int rows){
  PWCM array = new WINCOLORMAP [rows];
  for (int i = 0; i < rows; i++ )
     ; // array [i] = 0.0;
  membyte += rows * sizeof(WINCOLORMAP);
  return array;
}
//####################################################################\\
void DeAllocWCM1D (PWCM & x){
  if (x != NULL) delete x;
  x = NULL;
}
//####################################################################\\
void ReAllocWCM1D (PWCM & x, int rows){
  DeAllocWCM1D(x);
  x = AllocWCM1D(rows);
}
//####################################################################\\
void ProcessWCM1D (PWCM& x, int rows, int what){
  switch(what){
  case STD_ALLOC:
    x = AllocWCM1D(rows);
    break;
  case STD_DEALLOC:
    DeAllocWCM1D(x);
    break;
  case STD_REALLOC:
    ReAllocWCM1D(x, rows);
    break;
  default:
    MessageBox(NULL, "ProcessWCM1D", "Error", MB_OK);
    break;
  }
}

//####################################################################\\

